@@ -845,21 +845,16 @@ vips_exif_set_string_encoding( ExifData *ed,
 	ExifEntry *entry, unsigned long component, const char *data )
 {
 	char *str;
+	char *ascii;
 	int len;
 
 	str = drop_tail( data );
 
 	/* libexif can only really save ASCII to things like UserComment.
 	 */
-#ifdef HAVE_G_STR_TO_ASCII
-{
-	char *ascii;
-
 	ascii = g_str_to_ascii( str, NULL );
 	g_free( str );
 	str = ascii;
-}
-#endif /*HAVE_G_STR_TO_ASCII*/
 
 	/* libexif comment strings are not NULL-terminated, and have an 
 	 * encoding tag (always ASCII) in the first 8 bytes.
@@ -880,21 +875,16 @@ vips_exif_set_string_ascii( ExifData *ed,
 	ExifEntry *entry, unsigned long component, const char *data )
 {
 	char *str;
+	char *ascii;
 	int len;
 
 	str = drop_tail( data );
 
 	/* libexif can only really save ASCII to things like UserComment.
 	 */
-#ifdef HAVE_G_STR_TO_ASCII
-{
-	char *ascii;
-
 	ascii = g_str_to_ascii( str, NULL );
 	g_free( str );
 	str = ascii;
-}
-#endif /*HAVE_G_STR_TO_ASCII*/
 
 	/* ASCII strings are NULL-terminated.
 	 */
