@@ -20,6 +20,8 @@
  * 	- deprecate autorotate -- it's too difficult to support properly
  * 31/7/20
  * 	- block broken thumbnails, if we can
+ * 14/2/21 kleisauke
+ * 	- move GObject part to heif2vips.c
  */
 
 /*
@@ -72,18 +74,8 @@
  */
 #if defined(HAVE_HEIF_DECODER) || defined(HAVE_HEIF_ENCODER)
 
-#include <libheif/heif.h>
-
 #include "pforeign.h"
 
-void
-vips__heif_error( struct heif_error *error )
-{
-	if( error->code ) 
-		vips_error( "heif", "%s (%d.%d)", error->message, error->code,
-			error->subcode );
-}
-
 const char *vips__heif_suffs[] = { 
 	".heic",
 	".heif",
@@ -91,10 +83,12 @@ const char *vips__heif_suffs[] = {
 	NULL 
 };
 
-#endif /*defined(DECODE) || defined(ENCODE)*/
+#endif /*defined(HAVE_HEIF_DECODER) || defined(HAVE_HEIF_ENCODER)*/
 
 #ifdef HAVE_HEIF_DECODER
 
+#include <libheif/heif.h>
+
 #define VIPS_TYPE_FOREIGN_LOAD_HEIF (vips_foreign_load_heif_get_type())
 #define VIPS_FOREIGN_LOAD_HEIF( obj ) \
 	(G_TYPE_CHECK_INSTANCE_CAST( (obj), \
@@ -196,6 +190,14 @@ typedef struct _VipsForeignLoadHeif {
 
 } VipsForeignLoadHeif;
 
+void
+vips__heif_error( struct heif_error *error )
+{
+	if( error->code ) 
+		vips_error( "heif", "%s (%d.%d)", error->message, error->code,
+			error->subcode );
+}
+
 typedef struct _VipsForeignLoadHeifClass {
 	VipsForeignLoadClass parent_class;
 
@@ -230,7 +232,8 @@ vips_foreign_load_heif_build( VipsObject *object )
 	printf( "vips_foreign_load_heif_build:\n" );
 #endif /*DEBUG*/
 
-	if( vips_source_rewind( heif->source ) )
+	if( heif->source &&
+		vips_source_rewind( heif->source ) )
 		return( -1 );
 
 	if( !heif->ctx ) {
@@ -245,7 +248,6 @@ vips_foreign_load_heif_build( VipsObject *object )
 		}
 	}
 
-
 	if( VIPS_OBJECT_CLASS( vips_foreign_load_heif_parent_class )->
 		build( object ) )
 		return( -1 );
@@ -253,7 +255,6 @@ vips_foreign_load_heif_build( VipsObject *object )
 	return( 0 );
 }
 
-
 static const char *heif_magic[] = {
 	"ftypheic",	/* A regular heif image */
 	"ftypheix",	/* Extended range (>8 bit) image */
@@ -436,6 +437,7 @@ vips_foreign_load_heif_set_header( VipsForeignLoadHeif *heif, VipsImage *out )
 	heif_item_id id[16];
 	int n_metadata;
 	struct heif_error error;
+	VipsForeignHeifCompression compression;
 
 	/* We take the metadata from the non-thumbnail first page. HEIC 
 	 * thumbnails don't have metadata.
@@ -586,6 +588,29 @@ vips_foreign_load_heif_set_header( VipsForeignLoadHeif *heif, VipsImage *out )
 		vips_image_set_int( out, 
 			VIPS_META_PAGE_HEIGHT, heif->page_height );
 
+	/* Determine compression from HEIF "brand". heif_avif and heif_avis
+	 * were added in v1.7.
+	 */
+	compression = VIPS_FOREIGN_HEIF_COMPRESSION_HEVC;
+
+#ifdef HAVE_HEIF_AVIF
+{
+	const unsigned char *brand_data;
+
+	if( (brand_data = vips_source_sniff( heif->source, 12 )) ) {
+		enum heif_brand brand;
+		brand = heif_main_brand( brand_data, 12 );
+		if( brand == heif_avif || 
+			brand == heif_avis )
+			compression = VIPS_FOREIGN_HEIF_COMPRESSION_AV1;
+	}
+}
+#endif /*HAVE_HEIF_AVIF*/
+
+	vips_image_set_string( out, "heif-compression",
+		vips_enum_nick( VIPS_TYPE_FOREIGN_HEIF_COMPRESSION,
+			compression ) );
+
 	/* FIXME .. we always decode to RGB in generate. We should check for
 	 * all grey images, perhaps. 
 	 */
@@ -707,6 +732,9 @@ vips_foreign_load_heif_header( VipsForeignLoad *load )
 	}
 
 #ifdef DEBUG
+	printf( "page_width = %d\n", heif->page_width );
+	printf( "page_height = %d\n", heif->page_height );
+
 	printf( "n_top = %d\n", heif->n_top );
 	for( i = 0; i < heif->n_top; i++ ) {
 		printf( "  id[%d] = %d\n", i, heif->id[i] );
@@ -739,6 +767,50 @@ vips_foreign_load_heif_header( VipsForeignLoad *load )
 	return( 0 );
 }
 
+#ifdef DEBUG
+void
+vips__heif_image_print( struct heif_image *img )
+{
+	const static enum heif_channel channel[] = {
+		heif_channel_Y,
+		heif_channel_Cb,
+		heif_channel_Cr,
+		heif_channel_R,
+		heif_channel_G,
+		heif_channel_B,
+		heif_channel_Alpha,
+		heif_channel_interleaved
+	};
+
+	const static char *channel_name[] = {
+		"heif_channel_Y",
+		"heif_channel_Cb",
+		"heif_channel_Cr",
+		"heif_channel_R",
+		"heif_channel_G",
+		"heif_channel_B",
+		"heif_channel_Alpha",
+		"heif_channel_interleaved"
+	};
+
+	int i;
+
+	printf( "vips__heif_image_print:\n" );
+	for( i = 0; i < VIPS_NUMBER( channel ); i++ ) {
+		if( !heif_image_has_channel( img, channel[i] ) )
+			continue;
+
+		printf( "\t%s:\n", channel_name[i] ); 
+		printf( "\t\twidth = %d\n", 
+			heif_image_get_width( img, channel[i] ) );
+		printf( "\t\theight = %d\n", 
+			heif_image_get_height( img, channel[i] ) );
+		printf( "\t\tbits = %d\n", 
+			heif_image_get_bits_per_pixel( img, channel[i] ) );
+	}
+}
+#endif /*DEBUG*/
+
 static int
 vips_foreign_load_heif_generate( VipsRegion *or, 
 	void *seq, void *a, void *b, gboolean *stop )
@@ -782,48 +854,7 @@ vips_foreign_load_heif_generate( VipsRegion *or,
 		}
 
 #ifdef DEBUG
-{
-		const static enum heif_channel channel[] = {
-			heif_channel_Y,
-			heif_channel_Cb,
-			heif_channel_Cr,
-			heif_channel_R,
-			heif_channel_G,
-			heif_channel_B,
-			heif_channel_Alpha,
-			heif_channel_interleaved
-		};
-
-		const static char *channel_name[] = {
-			"heif_channel_Y",
-			"heif_channel_Cb",
-			"heif_channel_Cr",
-			"heif_channel_R",
-			"heif_channel_G",
-			"heif_channel_B",
-			"heif_channel_Alpha",
-			"heif_channel_interleaved"
-		};
-
-		int i;
-
-		printf( "vips_foreign_load_heif_generate:\n" );
-		for( i = 0; i < VIPS_NUMBER( channel ); i++ ) {
-			printf( "\t%s:\n", channel_name[i] ); 
-			printf( "\t\twidth = %d\n", 
-				heif_image_get_width( heif->img, 
-					channel[i] ) );
-			printf( "\t\theight = %d\n", 
-				heif_image_get_height( heif->img, 
-					channel[i] ) );
-			printf( "\t\tbits = %d\n", 
-				heif_image_get_bits_per_pixel( heif->img, 
-					channel[i] ) );
-			printf( "\t\thas_channel = %d\n", 
-				heif_image_has_channel( heif->img, 
-					channel[i] ) );
-		}
-}
+		vips__heif_image_print( heif->img );
 #endif /*DEBUG*/
 	}
 
@@ -1256,120 +1287,5 @@ vips_foreign_load_heif_source_init( VipsForeignLoadHeifSource *source )
 
 #endif /*HAVE_HEIF_DECODER*/
 
-/**
- * vips_heifload:
- * @filename: file to load
- * @out: (out): decompressed image
- * @...: %NULL-terminated list of optional named arguments
- *
- * Optional arguments:
- *
- * * @page: %gint, page (top-level image number) to read
- * * @n: %gint, load this many pages
- * * @thumbnail: %gboolean, fetch thumbnail instead of image
- *
- * Read a HEIF image file into a VIPS image. 
- *
- * Use @page to select a page to render, numbering from zero. If neither @n
- * nor @page are set, @page defaults to the primary page, otherwise to 0.
- *
- * Use @n to select the number of pages to render. The default is 1. Pages are
- * rendered in a vertical column. Set to -1 to mean "until the end of the 
- * document". Use vips_grid() to reorganise pages.
- *
- * HEIF images have a primary image. The metadata item `heif-primary` gives 
- * the page number of the primary.
- *
- * If @thumbnail is %TRUE, then fetch a stored thumbnail rather than the
- * image.
- *
- * See also: vips_image_new_from_file().
- *
- * Returns: 0 on success, -1 on error.
+/* The C API wrappers are defined in foreign.c.
  */
-int
-vips_heifload( const char *filename, VipsImage **out, ... )
-{
-	va_list ap;
-	int result;
-
-	va_start( ap, out );
-	result = vips_call_split( "heifload", ap, filename, out );
-	va_end( ap );
-
-	return( result );
-}
-
-/**
- * vips_heifload_buffer:
- * @buf: (array length=len) (element-type guint8): memory area to load
- * @len: (type gsize): size of memory area
- * @out: (out): image to write
- * @...: %NULL-terminated list of optional named arguments
- *
- * Optional arguments:
- *
- * * @page: %gint, page (top-level image number) to read
- * * @n: %gint, load this many pages
- * * @thumbnail: %gboolean, fetch thumbnail instead of image
- *
- * Read a HEIF image file into a VIPS image. 
- * Exactly as vips_heifload(), but read from a memory buffer. 
- *
- * You must not free the buffer while @out is active. The 
- * #VipsObject::postclose signal on @out is a good place to free. 
- *
- * See also: vips_heifload().
- *
- * Returns: 0 on success, -1 on error.
- */
-int
-vips_heifload_buffer( void *buf, size_t len, VipsImage **out, ... )
-{
-	va_list ap;
-	VipsBlob *blob;
-	int result;
-
-	/* We don't take a copy of the data or free it.
-	 */
-	blob = vips_blob_new( NULL, buf, len );
-
-	va_start( ap, out );
-	result = vips_call_split( "heifload_buffer", ap, blob, out );
-	va_end( ap );
-
-	vips_area_unref( VIPS_AREA( blob ) );
-
-	return( result );
-}
-
-/**
- * vips_heifload_source:
- * @source: source to load from
- * @out: (out): image to write
- * @...: %NULL-terminated list of optional named arguments
- *
- * Optional arguments:
- *
- * * @page: %gint, page (top-level image number) to read
- * * @n: %gint, load this many pages
- * * @thumbnail: %gboolean, fetch thumbnail instead of image
- *
- * Exactly as vips_heifload(), but read from a source. 
- *
- * See also: vips_heifload().
- *
- * Returns: 0 on success, -1 on error.
- */
-int
-vips_heifload_source( VipsSource *source, VipsImage **out, ... )
-{
-	va_list ap;
-	int result;
-
-	va_start( ap, out );
-	result = vips_call_split( "heifload_source", ap, source, out );
-	va_end( ap );
-
-	return( result );
-}
