@@ -64,7 +64,8 @@ int vips__tiff_write( VipsImage *in, const char *filename,
 	int level, 
 	gboolean lossless,
 	VipsForeignDzDepth depth,
-	gboolean subifd );
+	gboolean subifd, 
+	gboolean premultiply );
 
 int vips__tiff_write_buf( VipsImage *in, 
 	void **obuf, size_t *olen, 
@@ -83,7 +84,8 @@ int vips__tiff_write_buf( VipsImage *in,
 	int level, 
 	gboolean lossless,
 	VipsForeignDzDepth depth,
-	gboolean subifd );
+	gboolean subifd,
+	gboolean premultiply );
 
 gboolean vips__istiff_source( VipsSource *source );
 gboolean vips__istifftiled_source( VipsSource *source );
@@ -168,7 +170,7 @@ int vips__jpeg_write_target( VipsImage *in, VipsTarget *target,
 	gboolean optimize_coding, gboolean progressive, gboolean strip,
 	gboolean trellis_quant, gboolean overshoot_deringing,
 	gboolean optimize_scans, int quant_table,
-	VipsForeignJpegSubsample subsample_mode );
+	VipsForeignSubsample subsample_mode );
 
 int vips__jpeg_read_source( VipsSource *source, VipsImage *out,
 	gboolean header_only, int shrink, int fail, gboolean autorotate );
@@ -213,23 +215,11 @@ int vips__webp_write_target( VipsImage *image, VipsTarget *target,
 	gboolean min_size, int kmin, int kmax,
 	gboolean strip, const char *profile );
 
-int vips__openslide_isslide( const char *filename );
-int vips__openslide_read_header( const char *filename, VipsImage *out, 
-	int level, gboolean autocrop, 
-	char *associated, gboolean attach_associated );
-int vips__openslide_read( const char *filename, VipsImage *out, 
-	int level, gboolean autocrop, gboolean attach_associated );
-int vips__openslide_read_associated( const char *filename, VipsImage *out, 
-	const char *associated );
-
-gboolean vips_foreign_load_pdf_is_a_buffer( const void *buf, size_t len );
-gboolean vips_foreign_load_pdf_is_a( const char *filename );
-
 int vips__quantise_image( VipsImage *in, 
 	VipsImage **index_out, VipsImage **palette_out,
 	int colours, int Q, double dither );
 
-extern const char *vips__nifti_suffs[];
+extern const char *vips_foreign_nifti_suffs[];
 
 VipsBandFormat vips__foreign_nifti_datatype2BandFmt( int datatype );
 int vips__foreign_nifti_BandFmt2datatype( VipsBandFormat fmt );
@@ -239,8 +229,18 @@ typedef void *(*VipsNiftiMapFn)( const char *name, GValue *value, glong offset,
 void *vips__foreign_nifti_map( VipsNiftiMapFn fn, void *a, void *b );
 
 extern const char *vips__heif_suffs[];
-struct heif_error;
-void vips__heif_error( struct heif_error *error );
+
+extern const char *vips__jp2k_suffs[];
+int vips__foreign_load_jp2k_decompress( VipsImage *out,
+	int width, int height, gboolean ycc_to_rgb,
+	void *from, size_t from_length,
+	void *to, size_t to_length );
+int vips__foreign_load_jp2k_compress( VipsRegion *region,
+	VipsRect *tile, VipsTarget *target,
+	int tile_width, int tile_height,
+	gboolean save_as_ycc, gboolean subsample, gboolean lossless, int Q );
+
+extern const char *vips__jxl_suffs[];
 
 #ifdef __cplusplus
 }
