@@ -201,6 +201,8 @@
  * 	- add subifd
  * 6/6/20 MathemanFlo
  * 	- support 2 and 4 bit greyscale load
+ * 27/3/21
+ * 	- add jp2k decompresion
  */
 
 /*
@@ -253,6 +255,23 @@
 #include "pforeign.h"
 #include "tiff.h"
 
+/* Aperio TIFFs (svs) use these compression types for jp2k-compressed tiles.
+ */
+#define JP2K_YCC 33003
+#define JP2K_RGB 33005
+
+/* Bioformats uses this tag for jp2k compressed tiles.
+ */
+#define JP2K_LOSSY 33004
+
+/* Compression types we handle ourselves.
+ */
+static int rtiff_we_decompress[] = {
+	JP2K_YCC,
+	JP2K_RGB,
+	JP2K_LOSSY
+};
+
 /* What we read from the tiff dir to set our read strategy. For multipage
  * read, we need to read and compare lots of these, so it needs to be broken
  * out as a separate thing.
@@ -308,7 +327,7 @@ typedef struct _RtiffHeader {
 	 */
 	double stonits;
 
-	/* Number of subifds, if any.
+	/* Number of subifds, 0 for none.
 	 */
 	int subifd_count;
 
@@ -316,6 +335,11 @@ typedef struct _RtiffHeader {
 	 */
 	char *image_description;
 
+	/* TRUE if the compression type is not supported by libtiff directly
+	 * and we must read the raw data and decompress ourselves.
+	 */
+	gboolean we_decompress;
+
 } RtiffHeader;
 
 /* Scanline-type process function.
@@ -371,6 +395,12 @@ typedef struct _Rtiff {
 	 */
 	tdata_t contig_buf;
 
+	/* If we are decompressing, we need a buffer to read the raw tile to
+	 * before running the decompressor.
+	 */
+	tdata_t compressed_buf;
+	tsize_t compressed_buf_length;
+
 	/* The Y we are reading at. Used to verify strip read is sequential.
 	 */
 	int y_pos;
@@ -654,7 +684,7 @@ rtiff_set_page( Rtiff *rtiff, int page )
 		}
 
 		if( rtiff->subifd >= 0 ) {
-			int subifd_count;
+			uint16 subifd_count;
 			toff_t *subifd_offsets;
 
 			if( !TIFFGetField( rtiff->tiff, TIFFTAG_SUBIFD, 
@@ -664,8 +694,7 @@ rtiff_set_page( Rtiff *rtiff, int page )
 				return( -1 );
 			}
 
-			if( subifd_count <= 0 || 
-				rtiff->subifd >= subifd_count ) {
+			if( rtiff->subifd >= subifd_count ) {
 				vips_error( "tiff2vips", 
 					_( "subifd %d out of range, "
 						"only 0-%d available" ),
@@ -1720,12 +1749,47 @@ static int
 rtiff_read_tile( Rtiff *rtiff, tdata_t *buf, int x, int y )
 {
 #ifdef DEBUG_VERBOSE
-	printf( "rtiff_read_tile: x = %d, y = %d\n", x, y ); 
+	printf( "rtiff_read_tile: x = %d, y = %d, we_decompress = %d\n", 
+		x, y, rtiff->header.we_decompress ); 
 #endif /*DEBUG_VERBOSE*/
 
-	if( TIFFReadTile( rtiff->tiff, buf, x, y, 0, 0 ) < 0 ) { 
-		vips_foreign_load_invalidate( rtiff->out );
-		return( -1 ); 
+	if( rtiff->header.we_decompress ) {
+		ttile_t tile_no = TIFFComputeTile( rtiff->tiff, x, y, 0, 0 );
+
+		tsize_t size;
+
+		size = TIFFReadRawTile( rtiff->tiff, tile_no, 
+			rtiff->compressed_buf, rtiff->compressed_buf_length );
+		if( size <= 0 ) {
+			vips_foreign_load_invalidate( rtiff->out );
+			return( -1 ); 
+		}
+
+		switch( rtiff->header.compression ) {
+		case JP2K_YCC:
+		case JP2K_RGB:
+		case JP2K_LOSSY:
+			if( vips__foreign_load_jp2k_decompress( 
+				rtiff->out, 
+				rtiff->header.tile_width, 
+				rtiff->header.tile_height,
+				TRUE,
+				rtiff->compressed_buf, size,
+				buf, rtiff->header.tile_size ) ) 
+				return( -1 );
+			break;
+
+		default:
+			g_assert_not_reached();
+			break;
+		}
+
+	}
+	else {
+		if( TIFFReadTile( rtiff->tiff, buf, x, y, 0, 0 ) < 0 ) { 
+			vips_foreign_load_invalidate( rtiff->out );
+			return( -1 ); 
+		}
 	}
 
 	return( 0 ); 
@@ -1982,6 +2046,20 @@ rtiff_read_tilewise( Rtiff *rtiff, VipsImage *out )
 		return( -1 );
 	}
 
+	/* If we will be decompressing, we need a buffer large enough to hold
+	 * the largest compressed tile in any page.
+	 *
+	 * Allocate a buffer 2x the uncompressed tile size ... much simpler
+	 * than searching every page for the largest tile with
+	 * TIFFTAG_TILEBYTECOUNTS.
+	 */
+	if( rtiff->header.we_decompress ) {
+		rtiff->compressed_buf_length = 2 * rtiff->header.tile_size;
+		if( !(rtiff->compressed_buf = vips_malloc( VIPS_OBJECT( out ), 
+			rtiff->compressed_buf_length )) )
+			return( -1 );
+	}
+
 	/* Read to this image, then cache to out, see below.
 	 */
 	t[0] = vips_image_new(); 
@@ -2355,8 +2433,10 @@ rtiff_read_stripwise( Rtiff *rtiff, VipsImage *out )
 static int
 rtiff_header_read( Rtiff *rtiff, RtiffHeader *header )
 {
+	int i;
 	uint16 extra_samples_count;
 	uint16 *extra_samples_types;
+	uint16 subifd_count;
 	toff_t *subifd_offsets;
 	char *image_description;
 
@@ -2377,6 +2457,19 @@ rtiff_header_read( Rtiff *rtiff, RtiffHeader *header )
 	TIFFGetFieldDefaulted( rtiff->tiff, 
 		TIFFTAG_COMPRESSION, &header->compression );
 
+	/* One of the types we decompress?
+	 */
+	for( i = 0; i < VIPS_NUMBER( rtiff_we_decompress ); i++ )
+		if( header->compression == rtiff_we_decompress[i] ) {
+#ifdef DEBUG
+			printf( "rtiff_header_read: "
+				"compression %d handled by us\n", 
+				header->compression );
+#endif /*DEBUG*/
+			header->we_decompress = TRUE;
+			break;
+		}
+
 	/* We must set this here since it'll change the value of scanline_size.
 	 */
 	rtiff_set_decode_format( rtiff );
@@ -2422,9 +2515,9 @@ rtiff_header_read( Rtiff *rtiff, RtiffHeader *header )
 	/* Arbitrary sanity-checking limits.
 	 */
 	if( header->width <= 0 ||
-		header->width > VIPS_MAX_COORD ||
+		header->width >= VIPS_MAX_COORD ||
 		header->height <= 0 ||
-		header->height > VIPS_MAX_COORD ) {
+		header->height >= VIPS_MAX_COORD ) {
 		vips_error( "tiff2vips",
 			"%s", _( "width/height out of range" ) );
 		return( -1 );
@@ -2452,10 +2545,11 @@ rtiff_header_read( Rtiff *rtiff, RtiffHeader *header )
 			header->separate = TRUE; 
 	}
 
-	/* Stays zero if there's no SUBIFD.
+	/* TIFFGetField needs a uint16 to write count to.
 	 */
-	TIFFGetField( rtiff->tiff, TIFFTAG_SUBIFD, 
-		&header->subifd_count, &subifd_offsets );
+	if( TIFFGetField( rtiff->tiff, TIFFTAG_SUBIFD, 
+		&subifd_count, &subifd_offsets ) )
+		header->subifd_count = subifd_count;
 
 	/* IMAGEDESCRIPTION often has useful metadata. libtiff makes sure 
 	 * that data is null-terminated and contains no embedded null 
@@ -2471,6 +2565,8 @@ rtiff_header_read( Rtiff *rtiff, RtiffHeader *header )
 	 */
 	header->tiled = TIFFIsTiled( rtiff->tiff );
 
+
+
 #ifdef DEBUG
 	printf( "rtiff_header_read: header.width = %d\n", 
 		header->width );
@@ -2652,6 +2748,7 @@ rtiff_header_equal( RtiffHeader *h1, RtiffHeader *h2 )
 		h1->photometric_interpretation != 
 			h2->photometric_interpretation ||
 		h1->sample_format != h2->sample_format ||
+		h1->compression != h2->compression ||
 		h1->separate != h2->separate ||
 		h1->tiled != h2->tiled ||
 		h1->orientation != h2->orientation )
