@@ -108,15 +108,15 @@ extern char *vips__disc_threshold;
 extern gboolean vips__cache_dump;
 extern gboolean vips__cache_trace;
 
-extern int vips__n_active_threads;
-
 void vips__threadpool_init( void );
+void vips__threadpool_shutdown( void );
+int vips__thread_execute( const char *name, GFunc func, gpointer data );
 
 void vips__cache_init( void );
 
-void vips__print_renders( void );
-
-void vips__type_leak( void );
+int vips__print_renders( void );
+int vips__type_leak( void );
+int vips__object_leak( void );
 
 typedef int (*im__fftproc_fn)( VipsImage *, VipsImage *, VipsImage * );
 int im__fftproc( VipsImage *dummy, 
@@ -141,6 +141,7 @@ int vips_mapfilerw( VipsImage * );
 int vips_remapfilerw( VipsImage * );
 
 void vips__buffer_init( void );
+void vips__buffer_shutdown( void );
 
 void vips__copy_4byte( int swap, unsigned char *to, unsigned char *from );
 void vips__copy_2byte( gboolean swap, unsigned char *to, unsigned char *from );
@@ -237,7 +238,8 @@ int vips__byteswap_bool( VipsImage *in, VipsImage **out, gboolean swap );
 
 char *vips__xml_properties( VipsImage *image );
 
-void vips__cairo2rgba( guint32 *buf, int n );
+void vips__premultiplied_bgra2rgba( guint32 * restrict p, int n );
+void vips__bgra2rgba( guint32 * restrict p, int n );
 void vips__Lab2LabQ_vec( VipsPel *out, float *in, int width );
 void vips__LabQ2Lab_vec( float *out, VipsPel *in, int width );
 
