@@ -1,5 +1,4 @@
 # vim: set fileencoding=utf-8 :
-import filecmp
 import sys
 import os
 import shutil
@@ -11,14 +10,14 @@ from helpers import \
     JPEG_FILE, SRGB_FILE, MATLAB_FILE, PNG_FILE, TIF_FILE, OME_FILE, \
     ANALYZE_FILE, GIF_FILE, WEBP_FILE, EXR_FILE, FITS_FILE, OPENSLIDE_FILE, \
     PDF_FILE, SVG_FILE, SVGZ_FILE, SVG_GZ_FILE, GIF_ANIM_FILE, DICOM_FILE, \
-    BMP_FILE, NIFTI_FILE, ICO_FILE, HEIC_FILE, TRUNCATED_FILE, \
+    BMP_FILE, NIFTI_FILE, ICO_FILE, AVIF_FILE, TRUNCATED_FILE, \
     GIF_ANIM_EXPECTED_PNG_FILE, GIF_ANIM_DISPOSE_BACKGROUND_FILE, \
     GIF_ANIM_DISPOSE_BACKGROUND_EXPECTED_PNG_FILE, \
     GIF_ANIM_DISPOSE_PREVIOUS_FILE, \
     GIF_ANIM_DISPOSE_PREVIOUS_EXPECTED_PNG_FILE, \
     temp_filename, assert_almost_equal_objects, have, skip_if_no, \
     TIF1_FILE, TIF2_FILE, TIF4_FILE, WEBP_LOOKS_LIKE_SVG_FILE, \
-    WEBP_ANIMATED_FILE
+    WEBP_ANIMATED_FILE, JP2K_FILE, RGBA_FILE
 
 class TestForeign:
     tempdir = None
@@ -28,6 +27,7 @@ class TestForeign:
         cls.tempdir = tempfile.mkdtemp()
 
         cls.colour = pyvips.Image.jpegload(JPEG_FILE)
+        cls.rgba = pyvips.Image.new_from_file(RGBA_FILE)
         cls.mono = cls.colour.extract_band(1).copy()
         # we remove the ICC profile: the RGB one will no longer be appropriate
         cls.mono.remove("icc-profile-data")
@@ -38,12 +38,13 @@ class TestForeign:
         cls.cmyk.remove("icc-profile-data")
 
         im = pyvips.Image.new_from_file(GIF_FILE)
-        cls.onebit = im > 128
+        cls.onebit = im[1] > 128
 
     @classmethod
     def teardown_class(cls):
         shutil.rmtree(cls.tempdir, ignore_errors=True)
         cls.colour = None
+        cls.rgba = None
         cls.mono = None
         cls.rad = None
         cls.cmyk = None
@@ -75,7 +76,7 @@ class TestForeign:
         max_diff = (im - x).abs().max()
         assert max_diff == 0
 
-    def save_load_file(self, format, options, im, thresh):
+    def save_load_file(self, format, options, im, max_diff=0):
         # yuk!
         # but we can't set format parameters for pyvips.Image.new_temp_file()
         filename = temp_filename(self.tempdir, format)
@@ -86,8 +87,7 @@ class TestForeign:
         assert im.width == x.width
         assert im.height == x.height
         assert im.bands == x.bands
-        max_diff = (im - x).abs().max()
-        assert max_diff <= thresh
+        assert (im - x).abs().max() <= max_diff
         x = None
 
     def save_load_buffer(self, saver, loader, im, max_diff=0, **kwargs):
@@ -115,7 +115,7 @@ class TestForeign:
         assert (im - x).abs().max() <= max_diff
 
     def test_vips(self):
-        self.save_load_file(".v", "", self.colour, 0)
+        self.save_load_file(".v", "", self.colour)
 
         # check we can save and restore metadata
         filename = temp_filename(self.tempdir, ".v")
@@ -321,8 +321,8 @@ class TestForeign:
         self.save_load_buffer("pngsave_buffer", "pngload_buffer", self.colour)
         self.save_load("%s.png", self.mono)
         self.save_load("%s.png", self.colour)
-        self.save_load_file(".png", "[interlace]", self.colour, 0)
-        self.save_load_file(".png", "[interlace]", self.mono, 0)
+        self.save_load_file(".png", "[interlace]", self.colour)
+        self.save_load_file(".png", "[interlace]", self.mono)
 
         # size of a regular mono PNG 
         len_mono = len(self.mono.write_to_buffer(".png"))
@@ -389,32 +389,43 @@ class TestForeign:
         self.save_load("%s.tif", self.mono)
         self.save_load("%s.tif", self.colour)
         self.save_load("%s.tif", self.cmyk)
-
+        self.save_load("%s.tif", self.rgba)
         self.save_load("%s.tif", self.onebit)
-        self.save_load_file(".tif", "[bitdepth=1]", self.onebit, 0)
-        self.save_load_file(".tif", "[miniswhite]", self.onebit, 0)
-        self.save_load_file(".tif", "[bitdepth=1,miniswhite]", self.onebit, 0)
 
-        self.save_load_file(".tif",
-                            "[profile={0}]".format(SRGB_FILE),
-                            self.colour, 0)
-        self.save_load_file(".tif", "[tile]", self.colour, 0)
-        self.save_load_file(".tif", "[tile,pyramid]", self.colour, 0)
-        self.save_load_file(".tif", "[tile,pyramid,subifd]", self.colour, 0)
+        self.save_load_file(".tif", "[bitdepth=1]", self.onebit)
+        self.save_load_file(".tif", "[miniswhite]", self.onebit)
+        self.save_load_file(".tif", "[bitdepth=1,miniswhite]", self.onebit)
+
+        self.save_load_file(".tif", f"[profile={SRGB_FILE}]", self.colour)
+        self.save_load_file(".tif", "[tile]", self.colour)
+        self.save_load_file(".tif", "[tile,pyramid]", self.colour)
+        self.save_load_file(".tif", "[tile,pyramid,subifd]", self.colour)
         self.save_load_file(".tif",
                             "[tile,pyramid,compression=jpeg]", self.colour, 80)
         self.save_load_file(".tif",
                             "[tile,pyramid,subifd,compression=jpeg]", 
                             self.colour, 80)
-        self.save_load_file(".tif", "[bigtiff]", self.colour, 0)
+        self.save_load_file(".tif", "[bigtiff]", self.colour)
         self.save_load_file(".tif", "[compression=jpeg]", self.colour, 80)
         self.save_load_file(".tif",
                             "[tile,tile-width=256]", self.colour, 10)
 
         im = pyvips.Image.new_from_file(TIF2_FILE)
-        self.save_load_file(".tif", "[bitdepth=2]", im, 0)
+        self.save_load_file(".tif", "[bitdepth=2]", im)
         im = pyvips.Image.new_from_file(TIF4_FILE)
-        self.save_load_file(".tif", "[bitdepth=4]", im, 0)
+        self.save_load_file(".tif", "[bitdepth=4]", im)
+
+        filename = temp_filename(self.tempdir, '.tif')
+        self.colour.write_to_file(filename, pyramid=True, compression="jpeg")
+        x = pyvips.Image.new_from_file(filename, page=2)
+        assert x.width == 72
+        assert abs(x.avg() - 117.3) < 1
+
+        filename = temp_filename(self.tempdir, '.tif')
+        self.colour.write_to_file(filename, pyramid=True, subifd=True, compression="jpeg")
+        x = pyvips.Image.new_from_file(filename, subifd=1)
+        assert x.width == 72
+        assert abs(x.avg() - 117.3) < 1
 
         filename = temp_filename(self.tempdir, '.tif')
         x = pyvips.Image.new_from_file(TIF_FILE)
@@ -512,11 +523,17 @@ class TestForeign:
             buf2 = f.read()
         assert len(buf) == len(buf2)
 
+        filename = temp_filename(self.tempdir, '.tif')
+        self.rgba.write_to_file(filename, premultiply=True)
+        a = pyvips.Image.new_from_file(filename)
+        b = self.rgba.premultiply().cast("uchar").unpremultiply().cast("uchar")
+        assert (a == b).min() == 255
+
         a = pyvips.Image.new_from_buffer(buf, "", page=2)
         b = pyvips.Image.new_from_buffer(buf2, "", page=2)
         assert a.width == b.width
         assert a.height == b.height
-        assert a.avg() == b.avg()
+        assert (a == b).min() == 255
 
         # just 0/255 in each band, shrink with mode and all pixels should be 0
         # or 255 in layer 1
@@ -527,6 +544,16 @@ class TestForeign:
             z = y.hist_find(band=0)
             assert z(0, 0)[0] + z(255, 0)[0] == y.width * y.height
 
+    @skip_if_no("jp2kload")
+    @skip_if_no("tiffload")
+    def test_tiffjp2k(self):
+        self.save_load_file(".tif", "[tile,compression=jp2k]", self.colour, 80)
+        self.save_load_file(".tif",
+                            "[tile,pyramid,compression=jp2k]", self.colour, 80)
+        self.save_load_file(".tif",
+                            "[tile,pyramid,subifd,compression=jp2k]",
+                            self.colour, 80)
+
     @skip_if_no("magickload")
     def test_magickload(self):
         def bmp_valid(im):
@@ -775,6 +802,10 @@ class TestForeign:
 
         self.file_loader("openslideload", OPENSLIDE_FILE, openslide_valid)
 
+        source = pyvips.Source.new_from_file(OPENSLIDE_FILE)
+        x = pyvips.Image.new_from_source(source, "")
+        openslide_valid(x)
+
     @skip_if_no("pdfload")
     def test_pdfload(self):
         def pdf_valid(im):
@@ -801,61 +832,53 @@ class TestForeign:
     def test_gifload(self):
         def gif_valid(im):
             a = im(10, 10)
-            assert_almost_equal_objects(a, [33])
+            assert_almost_equal_objects(a, [33, 33, 33, 255])
             assert im.width == 159
             assert im.height == 203
-            assert im.bands == 1
+            assert im.bands == 3
 
         self.file_loader("gifload", GIF_FILE, gif_valid)
         self.buffer_loader("gifload_buffer", GIF_FILE, gif_valid)
 
-        # 'n' param added in 8.5
-        if pyvips.at_least_libvips(8, 5):
-            x1 = pyvips.Image.new_from_file(GIF_ANIM_FILE)
-            x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=2)
-            assert x2.height == 2 * x1.height
-            page_height = x2.get("page-height")
-            assert page_height == x1.height
+        # test metadata
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=-1)
+        # our test gif has delay 0 for the first frame set in error
+        assert x2.get("delay") == [0, 50, 50, 50, 50]
+        assert x2.get("loop") == 32760
+        assert x2.get("background") == [255, 255, 255]
+        # test deprecated fields too
+        assert x2.get("gif-loop") == 32759
+        assert x2.get("gif-delay") == 0
 
-            x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=-1)
-            assert x2.height == 5 * x1.height
-            # our test gif has delay 0 for the first frame set in error
-            assert x2.get("delay") == [0, 50, 50, 50, 50]
+        # test every pixel
+        x1 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=-1)
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_EXPECTED_PNG_FILE)
+        assert (x1 - x2).abs().max() == 0
 
-            x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, page=1, n=-1)
-            assert x2.height == 4 * x1.height
+        # test page handling
+        x1 = pyvips.Image.new_from_file(GIF_ANIM_FILE)
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=2)
+        assert x2.height == 2 * x1.height
+        page_height = x2.get("page-height")
+        assert page_height == x1.height
 
-            animation = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=-1)
-            filename = temp_filename(self.tempdir, '.png')
-            animation.write_to_file(filename)
-            # Uncomment to see output file
-            # animation.write_to_file('cogs.png')
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=-1)
+        assert x2.height == 5 * x1.height
 
-            assert filecmp.cmp(GIF_ANIM_EXPECTED_PNG_FILE, filename, shallow=False)
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, page=1, n=-1)
+        assert x2.height == 4 * x1.height
 
     @skip_if_no("gifload")
     def test_gifload_animation_dispose_background(self):
-        animation = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_BACKGROUND_FILE, n=-1)
-
-        filename = temp_filename(self.tempdir, '.png')
-        animation.write_to_file(filename)
-
-        # Uncomment to see output file
-        # animation.write_to_file('dispose-background.png')
-
-        assert filecmp.cmp(GIF_ANIM_DISPOSE_BACKGROUND_EXPECTED_PNG_FILE, filename, shallow=False)
+        x1 = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_BACKGROUND_FILE, n=-1)
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_BACKGROUND_EXPECTED_PNG_FILE)
+        assert (x1 - x2).abs().max() == 0
 
     @skip_if_no("gifload")
     def test_gifload_animation_dispose_previous(self):
-        animation = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_PREVIOUS_FILE, n=-1)
-
-        filename = temp_filename(self.tempdir, '.png')
-        animation.write_to_file(filename)
-
-        # Uncomment to see output file
-        # animation.write_to_file('dispose-previous.png')
-
-        assert filecmp.cmp(GIF_ANIM_DISPOSE_PREVIOUS_EXPECTED_PNG_FILE, filename, shallow=False)
+        x1 = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_PREVIOUS_FILE, n=-1)
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_PREVIOUS_EXPECTED_PNG_FILE)
+        assert (x1 - x2).abs().max() == 0
 
     @skip_if_no("svgload")
     def test_svgload(self):
@@ -884,6 +907,10 @@ class TestForeign:
         assert abs(im.width * 2 - x.width) < 2
         assert abs(im.height * 2 - x.height) < 2
 
+        with pytest.raises(pyvips.error.Error):
+            svg = b'<svg viewBox="0 0 0 0"></svg>'
+            im = pyvips.Image.new_from_buffer(svg, "")
+
     def test_csv(self):
         self.save_load("%s.csv", self.mono)
 
@@ -1072,37 +1099,47 @@ class TestForeign:
     def test_heifload(self):
         def heif_valid(im):
             a = im(10, 10)
-            # different versions of HEIC decode have slightly different 
+            # different versions of libheif decode have slightly different 
             # rounding
             assert_almost_equal_objects(a, [197.0, 181.0, 158.0], threshold=2)
             assert im.width == 3024
             assert im.height == 4032
             assert im.bands == 3
 
-        self.file_loader("heifload", HEIC_FILE, heif_valid)
-        self.buffer_loader("heifload_buffer", HEIC_FILE, heif_valid)
+        self.file_loader("heifload", AVIF_FILE, heif_valid)
+        self.buffer_loader("heifload_buffer", AVIF_FILE, heif_valid)
 
     @skip_if_no("heifsave")
     def test_heifsave(self):
         self.save_load_buffer("heifsave_buffer", "heifload_buffer",
-                              self.colour, 80)
-        self.save_load("%s.heic", self.colour)
+                              self.colour, 80, compression="av1")
+        # TODO: perhaps we should automatically set the compression to
+        # av1 when we save to *.avif?
+        #self.save_load("%s.avif", self.colour)
+        self.save_load_file(".avif", "[compression=av1]",
+                            self.colour, 80)
 
-        # test lossless mode
-        im = pyvips.Image.new_from_file(HEIC_FILE)
-        buf = im.heifsave_buffer(lossless=True)
-        im2 = pyvips.Image.new_from_buffer(buf, "")
+        # uncomment to test lossless mode, will take a while
+        #im = pyvips.Image.new_from_file(AVIF_FILE)
+        #buf = im.heifsave_buffer(lossless=True, compression="av1")
+        #im2 = pyvips.Image.new_from_buffer(buf, "")
         # not in fact quite lossless
-        assert abs(im.avg() - im2.avg()) < 3
+        #assert abs(im.avg() - im2.avg()) < 3
 
-        # higher Q should mean a bigger buffer
-        b1 = im.heifsave_buffer(Q=10)
-        b2 = im.heifsave_buffer(Q=90)
+        # higher Q should mean a bigger buffer, needs libheif >= v1.8.0,
+        # see: https://github.com/libvips/libvips/issues/1757
+        b1 = self.mono.heifsave_buffer(Q=10, compression="av1")
+        b2 = self.mono.heifsave_buffer(Q=90, compression="av1")
+        assert len(b2) > len(b1)
+
+        # Chroma subsampling should produce smaller file size for same Q
+        b1 = self.colour.heifsave_buffer(compression="av1", subsample_mode="on")
+        b2 = self.colour.heifsave_buffer(compression="av1", subsample_mode="off")
         assert len(b2) > len(b1)
 
         # try saving an image with an ICC profile and reading it back 
         # not all libheif have profile support, so put it in an if
-        buf = self.colour.heifsave_buffer()
+        buf = self.colour.heifsave_buffer(Q=10, compression="av1")
         im = pyvips.Image.new_from_buffer(buf, "")
         p1 = self.colour.get("icc-profile-data")
         if im.get_typeof("icc-profile-data") != 0:
@@ -1113,18 +1150,64 @@ class TestForeign:
         # the exif test will need us to be able to walk the header,
         # we can't just check exif-data
 
-        # libheif 1.1 (on ubuntu 18.04, current LTS) does not support exif
-        # write, so this test is commented out
-
         # test that exif changes change the output of heifsave
         # first make sure we have exif support
-        #z = pyvips.Image.new_from_file(JPEG_FILE)
-        #if z.get_typeof("exif-ifd0-Orientation") != 0:
-        #    x = self.colour.copy()
-        #    x.set("exif-ifd0-Make", "banana")
-        #    buf = x.heifsave_buffer()
-        #    y = pyvips.Image.new_from_buffer(buf, "")
-        #    assert y.get("exif-ifd0-Make").split(" ")[0] == "banana"
+        z = pyvips.Image.new_from_file(AVIF_FILE)
+        if z.get_typeof("exif-ifd0-Make") != 0:
+            x = z.copy()
+            x.set("exif-ifd0-Make", "banana")
+            buf = x.heifsave_buffer(Q=10, compression="av1")
+            y = pyvips.Image.new_from_buffer(buf, "")
+            assert y.get("exif-ifd0-Make").split(" ")[0] == "banana"
+
+    @skip_if_no("jp2kload")
+    def test_jp2kload(self):
+        def jp2k_valid(im):
+            a = im(402, 73)
+            assert_almost_equal_objects(a, [141, 144, 73], threshold=2)
+            assert im.width == 800
+            assert im.height == 400
+            assert im.bands == 3
+
+        self.file_loader("jp2kload", JP2K_FILE, jp2k_valid)
+        self.buffer_loader("jp2kload_buffer", JP2K_FILE, jp2k_valid)
+
+    @skip_if_no("jp2ksave")
+    def test_jp2ksave(self):
+        self.save_load_buffer("jp2ksave_buffer", "jp2kload_buffer",
+                              self.colour, 80)
+
+        buf = self.colour.jp2ksave_buffer(lossless=True)
+        im2 = pyvips.Image.new_from_buffer(buf, "")
+        assert (self.colour == im2).min() == 255
+
+        # higher Q should mean a bigger buffer
+        b1 = self.mono.jp2ksave_buffer(Q=10)
+        b2 = self.mono.jp2ksave_buffer(Q=90)
+        assert len(b2) > len(b1)
+
+        # disabling chroma subsample should mean a bigger buffer
+        b1 = self.colour.jp2ksave_buffer(subsample_mode="on")
+        b2 = self.colour.jp2ksave_buffer(subsample_mode="off")
+        assert len(b2) > len(b1)
+
+        # enabling lossless should mean a bigger buffer
+        b1 = self.colour.jp2ksave_buffer(lossless=False)
+        b2 = self.colour.jp2ksave_buffer(lossless=True)
+        assert len(b2) > len(b1)
+
+        # 16-bit colour load and save
+        im = self.colour.colourspace("rgb16")
+        buf = im.jp2ksave_buffer(lossless=True)
+        im2 = pyvips.Image.new_from_buffer(buf, "")
+        assert (im == im2).min() == 255
+
+        # openjpeg 32-bit load and save doesn't seem to work, comment out
+        # im = self.colour.colourspace("rgb16").cast("uint") << 14
+        # buf = im.jp2ksave_buffer(lossless=True)
+        # im2 = pyvips.Image.new_from_buffer(buf, "")
+        # assert (im == im2).min() == 255
+
 
 if __name__ == '__main__':
     pytest.main()
